"""
MCP23017 IO expander
"""

from __future__ import absolute_import

import logging
from typing import List, Optional, cast

from ...types import ConfigType, PinType
from . import GenericGPIO, InterruptEdge, InterruptSupport, PinDirection, PinPUD

_LOG = logging.getLogger(__name__)

REQUIREMENTS = ("adafruit_circuitpython_mcp230xx",)
CONFIG_SCHEMA = {
    "chip_addr": {"type": "integer", "required": False, "empty": False},
}


class GPIO(GenericGPIO):
    """
    Implementation of GPIO class for the MCP23017 IO expander chip.

    Pin numbers 0 - 15.
    """

    # | InterruptSupport.CAPTURE_REGISTER
    INTERRUPT_SUPPORT = (
        InterruptSupport.FLAG_REGISTER
        | InterruptSupport.INTERRUPT_PIN
        | InterruptSupport.SET_TRIGGERS
    )
    PIN_SCHEMA = {
        "pin": dict(type="integer", required=True, min=0, max=15),
    }

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        import board  # type: ignore
        import busio  # type: ignore
        import digitalio  # type: ignore
        from adafruit_mcp230xx import mcp23017  # type: ignore

        i2c = busio.I2C(board.SCL, board.SDA)
        self.direction_map = {
            PinDirection.INPUT: digitalio.Direction.INPUT,
            PinDirection.OUTPUT: digitalio.Direction.OUTPUT,
        }
        # Pulldowns are not supported on MCP23017
        self.pullup_map = {
            PinPUD.UP: digitalio.Pull.UP,
            PinPUD.OFF: None,
        }

        # Use the "protected" constant for the default address
        self.io = mcp23017.MCP23017(
            i2c,
            address=self.config.get(
                "chip_addr",
                mcp23017._MCP23017_ADDRESS,  # pylint: disable=protected-access
            ),
        )
        self.io.clear_ints()

    def setup_pin(
        self,
        pin: PinType,
        direction: PinDirection,
        pullup: PinPUD,
        pin_config: ConfigType,
        initial: Optional[str] = None,
    ) -> None:
        mcp_pin = self.io.get_pin(pin)
        mcp_pin.direction = self.direction_map[direction]
        if direction == PinDirection.OUTPUT:
            if initial is not None:
                mcp_pin.value = initial == "high"
        else:
            mcp_pin.pull = self.pullup_map[pullup]

    def setup_interrupt(
        self, pin: PinType, edge: InterruptEdge, in_conf: ConfigType
    ) -> None:
        # TODO: Tasks pending completion -@flyte at 29/01/2021, 19:42:23
        # Add ODR and INTPOL values of the IOCON register to the config for this module
        _LOG.debug(
            "MCP23017 module %s IOCON: %s", self.config["name"], bin(self.io.io_control)
        )

        # This is definitely an integer because of the PIN_SCHEMA for this class
        pin = cast(int, pin)

        if edge == InterruptEdge.BOTH:
            # Set this pin to interrupt when it changes from its previous value
            self.io.interrupt_configuration &= ~(1 << pin)
        else:
            # Enable comparison to default value for this pin
            self.io.interrupt_configuration |= 1 << pin

        if edge == InterruptEdge.RISING:
            # Set default value for this pin to be 0
            self.io.default_value &= ~(1 << pin)
        elif edge == InterruptEdge.FALLING:
            # Set default value for this pin to be 1
            self.io.default_value |= 1 << pin

        # Enable the interrupt on this pin
        self.io.interrupt_enable |= 1 << pin

    def set_pin(self, pin: PinType, value: bool) -> None:
        self.io.get_pin(pin).value = value

    def get_pin(self, pin: PinType) -> bool:
        return bool(self.io.get_pin(pin).value)

    def get_int_pins(self) -> List[PinType]:
        return cast(List[PinType], self.io.int_flag)



"""
Adafruit ADS1x15 analog to digital converters
"""

from typing import cast

from ...types import CerberusSchemaType, ConfigType, SensorValueType
from . import GenericSensor

SENSOR_ADS1015 = "ADS1015"
SENSOR_ADS1115 = "ADS1115"
SENSOR_TYPES = [SENSOR_ADS1015, SENSOR_ADS1115]

REQUIREMENTS = ("adafruit-circuitpython-ads1x15",)
CONFIG_SCHEMA: CerberusSchemaType = {
    "chip_addr": dict(type="integer", required=False, empty=False, default=0x48),
    "type": dict(
        type="string",
        required=True,
        empty=False,
        allowed=SENSOR_TYPES,
    ),
    "pin": dict(type="integer", required=True, empty=False, allowed=[0, 1, 2, 3]),
    "gain": dict(
        type="integer",
        required=False,
        empty=False,
        allowed=[2 / 3, 1, 2, 4, 8, 16],
        default=1,
    ),
}


class Sensor(GenericSensor):
    """
    Implementation of Sensor class for the Adafruit_ADS1x15.
    """

    SENSOR_SCHEMA: CerberusSchemaType = {
        "type": dict(
            type="string",
            required=False,
            empty=False,
            allowed=["value", "voltage"],
            default="value",
        ),
    }

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,attribute-defined-outside-init
        # pylint: disable=import-error,no-member
        import board  # type: ignore
        import busio  # type: ignore
        from adafruit_ads1x15.analog_in import AnalogIn  # type: ignore
        from adafruit_ads1x15.ads1x15 import ADS1x15  # type: ignore

        # Create the I2C bus
        self.i2c = busio.I2C(board.SCL, board.SDA)

        # Create the ADC object using the I2C bus
        if self.config["type"] == SENSOR_ADS1015:
            from adafruit_ads1x15.ads1015 import ADS1015  # type: ignore

            ad_sensor_class = ADS1015
        else:
            from adafruit_ads1x15.ads1115 import ADS1115  # type: ignore

            ad_sensor_class = ADS1115
        self.ads: ADS1x15 = ad_sensor_class(
            self.i2c, gain=self.config["gain"], address=self.config["chip_addr"]
        )

        # Create single-ended input on channel 0
        self.chan = AnalogIn(self.ads, self.config["pin"])

    def get_value(self, sens_conf: ConfigType) -> SensorValueType:
        """
        Get the value or voltage from the sensor
        """
        sens_type = sens_conf["type"]
        data = dict(value=self.chan.value, voltage=self.chan.voltage)
        return cast(
            float,
            data[sens_type],
        )



"""
BH1750 light level sensor
"""
from ...types import ConfigType, SensorValueType
from . import GenericSensor

REQUIREMENTS = ("smbus2",)

CONFIG_SCHEMA = {
    "i2c_bus_num": {"type": "integer", "required": True, "empty": False},
    "chip_addr": {"type": "integer", "required": True, "empty": False},
}

# Define some constants from the datasheet
DEVICE = 0x23  # Default device I2C address
POWER_DOWN = 0x00  # No active state
POWER_ON = 0x01  # Power on
RESET = 0x07  # Reset data register value
# Start measurement at 4lx resolution. Time typically 16ms.
CONTINUOUS_LOW_RES_MODE = 0x13
# Start measurement at 1lx resolution. Time typically 120ms
CONTINUOUS_HIGH_RES_MODE_1 = 0x10
# Start measurement at 0.5lx resolution. Time typically 120ms
CONTINUOUS_HIGH_RES_MODE_2 = 0x11
# Start measurement at 1lx resolution. Time typically 120ms
# Device is automatically set to Power Down after measurement.
ONE_TIME_HIGH_RES_MODE_1 = 0x20
# Start measurement at 0.5lx resolution. Time typically 120ms
# Device is automatically set to Power Down after measurement.
ONE_TIME_HIGH_RES_MODE_2 = 0x21
# Start measurement at 1lx resolution. Time typically 120ms
# Device is automatically set to Power Down after measurement.
ONE_TIME_LOW_RES_MODE = 0x23


class Sensor(GenericSensor):
    """
    Implementation of Sensor class for the BH1750 light sensor.
    """

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,attribute-defined-outside-init,import-error
        from smbus2 import SMBus  # type: ignore

        self.SMBus = SMBus  # pylint: disable=invalid-name
        self.bus_num: int = self.config["i2c_bus_num"]
        self.address: int = self.config["chip_addr"]

    def get_value(self, sens_conf: ConfigType) -> SensorValueType:
        """
        Get the light value from the sensor.
        """
        with self.SMBus(self.bus_num) as bus:
            data = bus.read_i2c_block_data(self.address, ONE_TIME_HIGH_RES_MODE_1, 2)
            result = (data[1] + (256 * data[0])) / 1.2
            return float(result)



"""
Contains the base class that is shared across all Stream modules.
"""

import asyncio
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

from ...types import ConfigType


class GenericStream(ABC):
    """
    Abstracts a generic stream interface to be implemented
    by the modules in this directory.
    """

    def __init__(self, config: ConfigType):
        self.config = config
        self.setup_module()

    @abstractmethod
    def setup_module(self) -> None:
        """
        Configure the module, open ports etc.
        """

    @abstractmethod
    def read(self) -> Optional[bytes]:
        """
        Read all available bytes from the stream. Return None if no bytes were available.
        """

    @abstractmethod
    def write(self, data: bytes) -> None:
        """
        Write bytes to the stream.
        """

    async def async_read(self) -> Optional[bytes]:
        """
        Use a ThreadPoolExecutor to call the module's synchronous read method.
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(ThreadPoolExecutor(), self.read)

    async def async_write(self, data: bytes) -> None:
        """
        Use a ThreadPoolExecutor to call the module's synchronous write method.
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(ThreadPoolExecutor(), self.write, data)

    def cleanup(self) -> None:
        """
        Called when closing the program to handle any cleanup operations.
        """



"""
Main entrypoint when MQTT IO is invoked as `python -m mqtt_io`.
"""
import argparse
import logging.config
import sys
from copy import deepcopy
from hashlib import sha256
from typing import Any, Optional

from mqtt_io.types import ConfigType

from . import VERSION
from .config import load_main_config
from .exceptions import ConfigValidationFailed
from .modules import install_missing_requirements
from .server import MqttIo


def hashed(value: Any) -> str:
    """
    Return the string representation of the value, hashed.
    """
    return sha256(str(value).encode("utf8")).hexdigest()


def redact_config(config: ConfigType) -> ConfigType:
    """
    Remove secret information from the config file.
    """
    ret = deepcopy(config)
    mqtt_config = config["mqtt"]
    ret["mqtt"]["host"] = hashed(mqtt_config["host"])
    if "password" in mqtt_config:
        ret["mqtt"]["password"] = "<provided but redacted>"
    for mqtt_section in ("host", "port", "user"):
        if mqtt_config[mqtt_section]:
            ret["mqtt"][mqtt_section] = hashed(mqtt_config[mqtt_section])
    return ret


def main() -> None:
    """
    Main entrypoint function.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("config")
    args = parser.parse_args()

    # Load, validate and normalise config, or quit.
    try:
        config = load_main_config(args.config)
    except ConfigValidationFailed as exc:
        print(str(exc), file=sys.stderr)
        sys.exit(1)

    if config["logging"]:
        logging.config.dictConfig(config["logging"])

    if config.get("reporting", {}).get("enabled"):
        # pylint: disable=import-outside-toplevel
        try:
            import sentry_sdk  # type: ignore
        except ImportError:
            install_missing_requirements(["sentry-sdk"])
            import sentry_sdk  # type: ignore

        issue_id: Optional[int] = config["reporting"].get("issue_id")

        sentry_sdk.init(
            "https://e3db7fd828ff468fb6caebd3953a69a2@o549418.ingest.sentry.io/5672194",
            traces_sample_rate=1.0,
            release=VERSION,
        )
        sentry_sdk.set_context("config", redact_config(config))
        if issue_id is not None:
            sentry_sdk.set_tag("issue_id", issue_id)

    mqtt_gpio = MqttIo(config)
    mqtt_gpio.run()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("")



"""
PCF8574 IO expander
"""

from typing import Optional, cast

from ...types import ConfigType, PinType
from . import GenericGPIO, PinDirection, PinPUD

REQUIREMENTS = ("pcf8574",)
CONFIG_SCHEMA = {
    "i2c_bus_num": {"type": "integer", "required": True, "empty": False},
    "chip_addr": {"type": "integer", "required": True, "empty": False},
}


class GPIO(GenericGPIO):
    """
    Implementation of GPIO class for the PCF8574 IO expander chip.
    """

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        # pylint: disable=no-name-in-module
        self.pullup_map = {PinPUD.UP: True, PinPUD.DOWN: False}
        from pcf8574 import PCF8574  # type: ignore

        self.io = PCF8574(self.config["i2c_bus_num"], self.config["chip_addr"])

    def setup_pin(
        self,
        pin: PinType,
        direction: PinDirection,
        pullup: PinPUD,
        pin_config: ConfigType,
        initial: Optional[str] = None,
    ) -> None:
        if direction == PinDirection.INPUT and pullup is not None:
            self.io.port[pin] = self.pullup_map[pullup]
        initial = pin_config.get("initial")
        if initial is not None:
            if initial == "high":
                self.set_pin(pin, True)
            elif initial == "low":
                self.set_pin(pin, False)

    def set_pin(self, pin: PinType, value: bool) -> None:
        self.io.port[pin] = value

    def get_pin(self, pin: PinType) -> bool:
        return cast(bool, self.io.port[pin])



"""
DHT11/DHT22/AM2302 temperature and humidity sensors
"""

from ...exceptions import RuntimeConfigError
from ...types import CerberusSchemaType, ConfigType, PinType, SensorValueType
from . import GenericSensor

REQUIREMENTS = ("Adafruit_DHT",)
ALLOWED_TYPES = ["dht11", "dht22", "am2302"]
CONFIG_SCHEMA: CerberusSchemaType = {
    "pin": dict(type="integer", required=True, empty=False),
    "type": dict(
        type="string",
        required=True,
        empty=False,
        allowed=ALLOWED_TYPES + [x.upper() for x in ALLOWED_TYPES],
    ),
}


class Sensor(GenericSensor):
    """
    Implementation of Sensor class for the DHT22 temperature sensor.
    """

    SENSOR_SCHEMA: CerberusSchemaType = {
        "type": dict(
            type="string",
            required=False,
            empty=False,
            default="temperature",
            allowed=["temperature", "humidity"],
        )
    }

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        import Adafruit_DHT as DHTsensor  # type: ignore

        sensor_type: str = self.config["type"].lower()

        self.sensor_type: int
        if sensor_type == "dht22":
            self.sensor_type = DHTsensor.DHT22
        elif sensor_type == "dht11":
            self.sensor_type = DHTsensor.DHT11
        elif sensor_type == "am2302":
            self.sensor_type = DHTsensor.AM2302
        else:
            raise RuntimeConfigError("Supported sensor types: DHT22, DHT11, AM2302")

        self.pin: PinType = self.config["pin"]
        self.sensor = DHTsensor

    def get_value(self, sens_conf: ConfigType) -> SensorValueType:
        """
        Get the temperature or humidity value from the sensor
        """
        humidity: SensorValueType
        temperature: SensorValueType
        humidity, temperature = self.sensor.read_retry(self.sensor_type, self.pin)
        if sens_conf["type"] == "temperature":
            return temperature
        if sens_conf["type"] == "humidity":
            return humidity
        raise RuntimeConfigError(
            "dht22 sensor '%s' was not configured to return 'temperature' or 'humidity'"
            % sens_conf["name"]
        )



"""
Home Assistant MQTT Discovery features.

https://www.home-assistant.io/docs/mqtt/discovery/
"""

import json
import logging
from typing import Any, Dict

from .constants import INPUT_TOPIC, OUTPUT_TOPIC, SENSOR_TOPIC, SET_SUFFIX
from .mqtt import MQTTClientOptions, MQTTMessageSend
from .types import ConfigType
from . import VERSION

_LOG = logging.getLogger(__name__)


def get_common_config(
    io_conf: ConfigType, mqtt_conf: ConfigType, mqtt_options: MQTTClientOptions
) -> Dict[str, Any]:
    """
    Return config that's common across all HQ discovery announcements.
    """
    disco_conf: ConfigType = mqtt_conf["ha_discovery"]
    config = dict(name=io_conf["name"])
    config.update(
        dict(
            availability_topic="/".join(
                (mqtt_conf["topic_prefix"], mqtt_conf["status_topic"])
            ),
            payload_available=mqtt_conf["status_payload_running"],
            payload_not_available=mqtt_conf["status_payload_dead"],
            device=dict(
                manufacturer="MQTT IO",
                model=f"v{VERSION}",
                identifiers=["mqtt-io", mqtt_options.client_id],
                name=disco_conf["name"],
            ),
        )
    )
    config.update(io_conf.get("ha_discovery", {}))
    return config


def hass_announce_digital_input(
    in_conf: ConfigType, mqtt_conf: ConfigType, mqtt_options: MQTTClientOptions
) -> MQTTMessageSend:
    """
    Create a message which announces digital input as binary_sensor to Home Assistant.
    """
    disco_conf: ConfigType = mqtt_conf["ha_discovery"]
    name: str = in_conf["name"]
    disco_prefix: str = disco_conf["prefix"]
    sensor_config = get_common_config(in_conf, mqtt_conf, mqtt_options)
    sensor_config.update(
        dict(
            unique_id=f"{mqtt_options.client_id}_{in_conf['module']}_input_{name}",
            state_topic="/".join((mqtt_conf["topic_prefix"], INPUT_TOPIC, name)),
            payload_on=in_conf["on_payload"],
            payload_off=in_conf["off_payload"],
        )
    )
    return MQTTMessageSend(
        "/".join(
            (
                disco_prefix,
                sensor_config.pop("component", "binary_sensor"),
                mqtt_options.client_id,
                name,
                "config",
            )
        ),
        json.dumps(sensor_config).encode("utf8"),
        retain=True,
    )


def hass_announce_digital_output(
    out_conf: ConfigType,
    mqtt_conf: ConfigType,
    mqtt_options: MQTTClientOptions,
) -> MQTTMessageSend:
    """
    Create a message which announces digital output as switch to Home Assistant.
    """
    disco_conf: ConfigType = mqtt_conf["ha_discovery"]
    name: str = out_conf["name"]
    prefix: str = mqtt_conf["topic_prefix"]
    disco_prefix: str = disco_conf["prefix"]
    switch_config = get_common_config(out_conf, mqtt_conf, mqtt_options)
    switch_config.update(
        dict(
            unique_id=f"{mqtt_options.client_id}_{out_conf['module']}_output_{name}",
            state_topic="/".join((prefix, OUTPUT_TOPIC, name)),
            command_topic="/".join((prefix, OUTPUT_TOPIC, name, SET_SUFFIX)),
            payload_on=out_conf["on_payload"],
            payload_off=out_conf["off_payload"],
        )
    )
    return MQTTMessageSend(
        "/".join(
            (
                disco_prefix,
                switch_config.pop("component", "switch"),
                mqtt_options.client_id,
                name,
                "config",
            )
        ),
        json.dumps(switch_config).encode("utf8"),
        retain=True,
    )


def hass_announce_sensor_input(
    sens_conf: ConfigType,
    mqtt_conf: ConfigType,
    mqtt_options: MQTTClientOptions,
) -> MQTTMessageSend:
    """
    Create a message which announces digital output as sensor to Home Assistant.
    """
    disco_conf: ConfigType = mqtt_conf["ha_discovery"]
    name: str = sens_conf["name"]
    prefix: str = mqtt_conf["topic_prefix"]
    disco_prefix: str = disco_conf["prefix"]
    sensor_config = get_common_config(sens_conf, mqtt_conf, mqtt_options)
    sensor_config.update(
        dict(
            unique_id=f"{mqtt_options.client_id}_{sens_conf['module']}_sensor_{name}",
            state_topic="/".join((prefix, SENSOR_TOPIC, name)),
        )
    )
    if "expire_after" not in sensor_config:
        sensor_config["expire_after"] = sens_conf["interval"] * 2 + 5

    return MQTTMessageSend(
        "/".join(
            (
                disco_prefix,
                sensor_config.pop("component", "sensor"),
                mqtt_options.client_id,
                name,
                "config",
            )
        ),
        json.dumps(sensor_config).encode("utf8"),
        retain=True,
    )



# pylint: disable-all

import os.path
from argparse import ArgumentParser
from shutil import move
from textwrap import dedent, indent
from typing import Any, Callable, List, Optional, TextIO

import md_toc  # type: ignore
import yaml

from ..types import ConfigType
from . import get_main_schema


def meta_entry(section: ConfigType, meta_key: str) -> Any:
    entry = section.get("meta", {}).get(meta_key, "")
    if hasattr(entry, "strip"):
        entry = entry.strip()
    return entry or None


def titleify(text: str) -> str:
    return text.replace("_", " ").title()


def format_yaml_value(value: Any) -> str:
    value_str: str = yaml.dump(value, default_style=None)
    if value_str.endswith("\n...\n"):
        value_str = " " + value_str[: -len("\n...\n")]
    elif "\n" not in value_str.rstrip("\n"):
        value_str = " " + value_str.rstrip("\n")
    else:
        value_str = "\n" + indent(value_str.rstrip("\n"), "  ")
    return value_str


class SectionDocumenter:
    def __init__(self, doc: Callable[[str], None]):
        self.doc = doc

    def document_schema_section(
        self, schema_section: ConfigType, parent_sections: Optional[List[str]] = None
    ) -> None:
        if parent_sections is None:
            parent_sections = []
        else:
            parent_sections = parent_sections.copy()

        child_schema = schema_section.get("schema")
        if child_schema:
            # This must be a list, which has a schema for each item
            parent_sections.append("*")
            self.document_schema_section(child_schema, parent_sections)
            return
        for entry_name in schema_section.keys():
            entry: ConfigType = schema_section[entry_name]
            self.document_cerberus_section(entry_name, entry, parent_sections)

    def document_cerberus_section(
        self, entry_name: str, cerberus_section: ConfigType, parent_sections: List[str]
    ) -> None:
        parent_sections = parent_sections.copy()

        auto_title = ""
        if parent_sections:
            # auto_title += "_" + ".".join(parent_sections) + "_."
            auto_title += ".".join(parent_sections) + "."
        auto_title += f"`{entry_name}`"

        title = meta_entry(cerberus_section, "title") or auto_title

        child_schema = cerberus_section.get("schema")

        depth = len([x for x in parent_sections if x != "*"])
        dashes = "- " * depth
        hashes = "#" * (depth + 1)
        if child_schema and not entry_name.endswith("*"):
            self.doc(f"\n\n{hashes} {title}")
        else:
            self.doc(f"\n\n{dashes}{hashes} {title}")

        self.doc("\n\n```yaml")

        type_str: str = format_yaml_value(cerberus_section["type"])
        self.doc(f"\nType:{type_str}")

        required: bool = cerberus_section.get("required", False)
        self.doc("\nRequired: %s" % ("yes" if required else "no"))

        unit = meta_entry(cerberus_section, "unit")
        if unit is not None:
            self.doc(f"\nUnit: {unit}")

        allowed = cerberus_section.get("allowed")
        if allowed is not None:
            allowed_str: str = yaml.dump(allowed)
            self.doc(f"\nAllowed:\n{allowed_str.strip()}")

        min_val = cerberus_section.get("min")
        if min_val is not None:
            self.doc(f"\nMinimum value: {min_val}")

        max_val = cerberus_section.get("max")
        if max_val is not None:
            self.doc(f"\nMinimum value: {max_val}")

        allow_unknown = cerberus_section.get("allow_unknown")
        if allow_unknown is not None:
            allow_unknown_str = "yes" if allow_unknown else "no"
            self.doc(f"\nUnlisted entries accepted: {allow_unknown_str}")

        if "default" in cerberus_section:
            default_str = format_yaml_value(cerberus_section["default"])
            self.doc(f"\nDefault:{default_str}")

        self.doc("\n```")

        description = meta_entry(cerberus_section, "description")
        if description:
            self.doc(f"\n\n{description}")

        extra_info = meta_entry(cerberus_section, "extra_info")
        if extra_info:
            self.doc(f"\n\n> {extra_info}")

        yaml_example = meta_entry(cerberus_section, "yaml_example")
        if yaml_example:
            self.doc(f"\n\n**Example:**\n\n```yaml\n{yaml_example}\n```")

        yaml_example_expand = meta_entry(cerberus_section, "yaml_example_expand")
        if yaml_example_expand:
            self.doc(
                dedent(
                    f"""\n
                    <details>
                    <summary>View example</summary>

                    ```yaml
                    {yaml_example_expand}
                    ```
                    </details>
                    """
                ).rstrip("\n")
            )

        if child_schema:
            parent_sections.append(entry_name)
            if "type" in child_schema:
                self.document_cerberus_section("*", child_schema, parent_sections)
            else:
                self.document_schema_section(child_schema, parent_sections)


def document_schema(doc_file: TextIO) -> None:
    schema = get_main_schema()

    def doc(entry: str) -> None:
        print(entry, file=doc_file, end="")

    doc(
        dedent(
            """
            The software is configured using a single YAML config file. This document details
            the config options for each section and provides examples.

            <!-- TOC -->
            """
        ).strip()
    )

    sd = SectionDocumenter(doc)
    sd.document_schema_section(schema)


def add_toc(path: str) -> None:
    toc = md_toc.build_toc(path, keep_header_levels=10, skip_lines=1)
    md_toc.write_string_on_file_between_markers(path, toc, "<!-- TOC -->")


def main() -> None:
    p = ArgumentParser()
    p.add_argument("output_path")
    args = p.parse_args()
    path, fname = os.path.split(args.output_path)
    tmp_path = os.path.join(path, f"~{fname}")
    with open(tmp_path, "w") as _doc_file:
        document_schema(_doc_file)
    add_toc(tmp_path)
    move(tmp_path, args.output_path)


if __name__ == "__main__":
    main()



"""
Abstraction layer for MQTT clients to enable easier switching out.
"""
import abc
import asyncio
import ssl
from dataclasses import dataclass
from enum import Enum, auto
from importlib import import_module
from typing import List, Optional, Tuple, Type


class MQTTProtocol(Enum):
    """
    MQTT protocol.
    """

    V31 = auto()
    V311 = auto()
    V5 = auto()


@dataclass
class MQTTMessage:
    """
    MQTT message.
    """

    topic: str
    payload: Optional[bytes] = None


@dataclass
class MQTTMessageSend(MQTTMessage):
    """
    MQTT message being sent.
    """

    qos: int = 0
    retain: bool = False


@dataclass
class MQTTWill:
    """
    MQTT will.
    """

    topic: str
    payload: Optional[bytes]
    qos: int
    retain: bool


@dataclass
class MQTTTLSOptions:
    """
    MQTT TLS options.
    """

    ca_certs: Optional[str] = None
    certfile: Optional[str] = None
    keyfile: Optional[str] = None
    cert_reqs: int = ssl.CERT_REQUIRED
    tls_version: int = ssl.PROTOCOL_TLS
    ciphers: Optional[str] = None

    @property
    def ssl_context(self) -> ssl.SSLContext:
        """
        The TLS options as an ssl.SSLContext object.
        """
        context = ssl.SSLContext(self.tls_version)

        if self.certfile is not None:
            context.load_cert_chain(self.certfile, self.keyfile)

        if self.cert_reqs == ssl.CERT_NONE:
            context.check_hostname = False

        context.verify_mode = self.cert_reqs or ssl.CERT_REQUIRED

        if self.ca_certs is None:
            context.load_default_certs()
        else:
            context.load_verify_locations(self.ca_certs)

        if self.ciphers is not None:
            context.set_ciphers(self.ciphers)

        return context


@dataclass
class MQTTClientOptions:  # pylint: disable=too-many-instance-attributes
    """
    MQTT client options.
    """

    hostname: str
    client_id: str
    port: int = 1883
    username: Optional[str] = None
    password: Optional[str] = None
    tls_options: Optional[MQTTTLSOptions] = None
    protocol: MQTTProtocol = MQTTProtocol.V311
    clean_session: Optional[bool] = None
    keepalive: int = 60
    will: Optional[MQTTWill] = None
    message_queue_size: int = 1000


class AbstractMQTTClient(abc.ABC):
    """
    Abstract MQTT client interface to enable easier switching out of MQTT client
    implementations.
    """

    @abc.abstractmethod
    def __init__(self, options: MQTTClientOptions):
        """
        Initialisation.
        """
        self._options = options

    @abc.abstractmethod
    async def connect(self, timeout: int = 10) -> None:
        """
        Connect to MQTT server.
        """

    @abc.abstractmethod
    async def disconnect(self) -> None:
        """
        Disconnect from MQTT server.
        """

    @abc.abstractmethod
    async def subscribe(self, topics: List[Tuple[str, int]]) -> None:
        """
        Subscribe to a list of topics.
        """

    @abc.abstractmethod
    async def publish(self, msg: MQTTMessageSend) -> None:
        """
        Publish an MQTT message.
        """

    @abc.abstractproperty
    def message_queue(self) -> "asyncio.Queue[MQTTMessage]":
        """
        Get the queue onto which received MQTT messages are put.
        """

    @staticmethod
    def get_implementation(module_name: str) -> Type["AbstractMQTTClient"]:
        """
        Import an implemenation of AbstractMQTTClient and return its class.
        """
        client_module = import_module(module_name)
        client: Type[AbstractMQTTClient]
        client = client_module.MQTTClient  # type: ignore[attr-defined]
        return client
